// main.h
#include "common.h"

const char version[64] = "\n\nJed Margolin's GPS Flash Erase   Version 1/17/2025\n";  // for uart
//================================================================
// Declare and initialize the switches
struct myswitch switch1 = {0xFF, 0x01, 0x01, 0x00};
struct myswitch switch2 = {0xFF, 0x01, 0x01, 0x00};
struct myswitch switch3 = {0xFF, 0x01, 0x01, 0x00};
struct myswitch switch4 = {0xFF, 0x01, 0x01, 0x00};

void do_switch(unsigned char, struct myswitch *);
//=======================================
extern void do_bme280(void);
extern void start_bme280(void);

extern unsigned char timer_flag;
//================================================
const struct messages dmsg1 = {0x80, 14,"Jed Margolin's"};
const struct messages dmsg2 = {0xC0, 15, "GPS Flash Erase"};
const struct messages dmsg3 = {0x83, 9, "1/17/2025"};
const struct messages dmsg4 = {0xC0, 6, "Erased"};
const struct messages dmsg5 = {0x80, 9, "Flash Log"};
const struct messages dmsg6 = {0x80, 10, "Start Date"};
const struct messages dmsg7 = {0x80, 14, "Barometric Ref"};
const struct messages dmsg8 = {0x80, 12, "Local Offset"};
const struct messages dmsg9 = {0x80, 9, "DST Dates"};
const struct messages dmsg10 = {0x80, 10, "Not Erased"};
const struct messages dmsg11 = {0x80, 5, "Ready"};
//=======================================================
void blink_led(uint16_t, uint16_t);
uint16_t mode;   // 1 = bme-280 and GPS; 0 = Set barometer, local offset, and DST
uint16_t secs_flag, irq_flag, irq_count, save_count; 
//=======================================================
// LCD Functions
void lcd_satinv(void);
extern void SetChrMode(void);
extern void SetCmdMode(void);
extern void lcd_init(void);
extern void domsg(const struct messages *);
extern void lcd_wr_data(unsigned char);
extern void lcd_wr_instr(unsigned char);
extern void lcd_text(char *);
extern void lcd_byte(char);
extern void lcd_wr_hex(unsigned int, unsigned char);
extern void lcd_wr_hex16(unsigned int, unsigned int);
extern void convert_hex_bcd2(unsigned char, unsigned char *);
extern void lcd_wr_bcd(unsigned int, unsigned char);
extern void lcd_wr_bcd2(unsigned int, unsigned char);
extern void lcd_wr_time(unsigned int, unsigned int);
extern void lcd_wr_hex16(unsigned int, unsigned int);
extern void clear_lcd(unsigned char);
//=======================================================
struct repeating_timer timer;
extern uint16_t logging;
void test_sint16(int8_t, int16_t);
void uint16_hex_bin_str(uint16_t, char *);
//=================================================================
char line[512];
char nstr[512];

extern void dump_flash(uint32_t, uint32_t);
extern void dump_flash_page_hex(uint32_t); 

void LED_break(void);
unsigned char uart_char, rxdata;
void echo_gps_msgs(void);       // For testing, just echo the GPS messages to the uart

extern uint16_t erase_flag;
void do_uart_RX(void );
extern void myflash_erase_sector(uint16_t);

extern uint16_t ts_number;
uint16_t data_count;

extern void uint8_bcd_str(unsigned char, char *);   // unsigned char to ASCII string, 3 digits  
extern void uint8_bcd_str2(unsigned char, char *);  // unsigned char to ASCII string, 2 digits
extern void sint8_bcd_str(int8_t, char *);          // signed char to ASCII string, 3 digits    
extern void uint16_bcd_str(uint16_t, char *);       // uint16 to ASCII string, 5 digits
extern void uint16_bcd_str4(uint16_t, char *);      // uint16 to ASCII string, 4 digits
extern void sint16_bcd_str(int16_t, char *);        // int16 to ASCII string, 5 digits
extern void uint32_bcd_str(uint32_t, char *);       // uint32_t to ASCII string, 8 digits
extern void sint32_bcd_str(int32_t, char *);        // int32_t to ASCII string, 8 digits
extern void uint8_hex_str(uint8_t, char *);         // unsigned char to ASCII string as hex digits, 2 digits    
extern void uint16_hex_str(uint16_t, char *);       // uint16 to ASCII string as 4 hex digits
extern void uint32_hex_str(unsigned long, char *);  // uint32 to ASCII string as 8 hex digits


void test_messages(void);
extern void myflash_erase_sector(uint16_t);
extern void erase_log_sectors(void);
//===================
/*  RPI Pico sizes
Name 	              C++      Bytes   Range

bool 		                     1 	    true or false
signed char 	     int8_t 	1 	    -128 to 127
unsigned char 	     uint8_t 	1 	    0 to 255
short int 	        int16_t 	2 	    -32768 to 32767
unsigned short int  uint16_t 	2 	    0 to 65535
int 	              int32_t 	4      -2147483648 to 2147483647
unsigned int 	     uint32_t 	4 	    0 to 4294967295
long int 	        int32_t 	4 	    -2147483648 to 2147483647
unsigned long int   uint32_t 	4 	    0 to 4294967295
long long 	        int64_t 	8 	    -9,223,372,036,854,775,808 to 9,223,372,036,854,775,807
unsigned long long  uint64_t 	8 	    0 to 18,446,744,073,709,551,615
float 		                  4 	    +/- 3.4e +/- 38 (~7 digits)
double 		                  8 	    +/- 1.7e +/- 308 (~15 digits)
long double 		            8 	    +/- 1.7e +/- 308 (~15 digits)
wchar_t 		             2 or 4      1 wide character
*/
//============================================
//***************************************************
/* Flash Memory
Flash Memory

There are 256 bytes in a Page.
There are 16 pages in a Sector (4096 bytes).
For Writing you Write a Page.
For Erasing you Erase a Sector.
Writing and Erasing are referenced to Flash Memory so it starts at Flash Memory address = 0.

#define FLASH_PAGE_SIZE 256      // for writing, already defined
#define FLASH_SECTOR_SIZE 4096   // for erasing Flash, already defined

For my Flash Logs I am allowing 512K for program so for Writing and Erasing, my data starts 
at 0x0008 0000 (524,288) 

#define FLASH_TARGET_ADDRESS (512 * 1024)  // 0x0008 0000 (524,288)    

A blank location is all ones. Writing can only change a '1' to a '0'. If you write 
a '1' you can write it again to change it to a '0'. 

The memory is rated for 100K endurance cycles. An endurance cycle is an erase and/or presumably a write. 
It isn't clear, and Winbond tech support doesn't understand the question of whether writes (only) count 
as an endurance cycle. I asked because flash memory has a charge pump to produce the voltages needed
to erase and write to the memory. That is why it can operate on a single supply (3.3V). Turning the
charge pump on probably produces some "wear". My question was if just writing uses up an endurance cycle. 
Winbond assumes you always erase a sector before writing to it even though you can separately write 
individual pages in a sector at different times. 

For writing a page: 
#define FLASH_PAGE_SIZE 256  
#define FLASH_TARGET_ADDRESS (512 * 1024)  //  524,288 = 0x0008 0000

Erasing a sector and writing a page uses the flash memory's address. If you use a Page_Index and/or a
Sector_Index you have to turn it into a Flash Memory Address.

//   flash_write_address = FLASH_TARGET_ADDRESS + (Flash_Page_Index * 256);  // There are 256 bytes in a page
   flash_write_address = FLASH_TARGET_ADDRESS + (Flash_Page_Index<<8);   // I do it with a shift instead of a multiply
   flash_program(flash_write_address, Data_Buffer, FLASH_PAGE_SIZE);     // Don't forget to put 256 bytes in the buffer
                                                                         // Use 0xFF for unused bytes
For erasing a Sector - I don't know what happens if you give it an address that is not on a 4096 byte boundary 
so I won't try it. 

#define FLASH_TARGET_ADDRESS (512 * 1024)    //  524,288 = 0x0008 0000
#define FLASH_SECTOR_SIZE 4096               // for erasing Flash

// Erase Address = FLASH_TARGET_ADDRESS + (My_Erase_Sector_Index * 4096); 
   Erase Address = FLASH_TARGET_ADDRESS + (My_Erase_Sector_Index<<12); 
   flash_range_erase(Erase_Address, FLASH_SECTOR_SIZE);  

These examples are in the part of flash memory that I use. They are for writing pages. FLASH_TARGET_ADDRESS is 512*1024

Page_Index       Flash Address
   0           FLASH_TARGET Address  + 0
   1           FLASH_TARGET Address  + 256
   2           FLASH_TARGET Address  + 512
   ...         ...
   16          FLASH_TARGET Address  + 4096

If you are using a Flash Sector_Index (for erasing)
Sector_Index       Flash Address
   0           FLASH_TARGET Address  + 0
   1           FLASH_TARGET Address  + 4096
   2           FLASH_TARGET Address  + 8192
   ...         ...
   to the end of Flash

Because I write only a single page at a time, every time I write 16 continguous pages I increment my Sector_Index (for erasing)

In some programs I use 1 MByte for a log (4096 sectors) and one or more seperate sectors for other things.

If the data in these separate sectors is less than 4096 bytes I just waste the remaining bytes. 
If I write a page of less than 256 bytes I waste the remaining bytes and write 0xFF to them.

These separate pages start at page 4096 (sector 256)

For example:

Separate Sectors
Separate Sector   Sector_Index                        Flash Memory Address                                   Page_Index
   Sep_Sec 0         256         FLASH_TARGET Address  + (256 * 4096) = FLASH_TARGET_ADDRESS + 1,048,576        4096
   Sep_Sec 1         257         FLASH_TARGET Address  + (257 * 4096) = FLASH_TARGET_ADDRESS + 1,052,672        4112
   Sep_Sec 2         258         FLASH_TARGET Address  + (258 * 4096) = FLASH_TARGET_ADDRESS + 1,056,768        4128
   Sep_Sec_3         259         FLASH_TARGET Address  + (259 * 4096) = FLASH_TARGET_ADDRESS + 1,060,864        4144
   Sep_Sec_4         260         FLASH_TARGET Address  + (260 * 4096) = FLASH_TARGET_ADDRESS + 1,064,960        4160

You can get the flash memory address from the Page Index but you run the risk of not having it 
aligned on a Sector address. I don't know what that would do other than confuse everything. 
You might erase a Page that you hadn't intended to.

Warning!  

If you are using core 1 it has to be turned off when writing or erasing flash memeory. 
Otherwse it may go off into the weeds. I use core 1 for pwm which uses interrupts. 
This is what works for me:

1.  writing a page (256 bytes)

void write_page(uint32_t page, unsigned char *buffer)
{
   uint32_t myflash_write_address;

   myflash_write_address = FLASH_TARGET_OFFSET + (page<<8); 

   ints = save_and_disable_interrupts();  

   multicore_reset_core1();
   multicore_fifo_pop_blocking();

   flash_range_program(myflash_write_address, buffer, FLASH_PAGE_SIZE); 

   restore_interrupts(ints);

   multicore_launch_core1(core1_entry);

   sleep_ms(10);   // Wait for it to start up    
 
//   show_page(page);
}


2.  Erasing a sector (4096 bytes)

void myflash_erase_sector(uint16_t sector)
{
// #define FLASH_TARGET_OFFSET (512 * 1024)  // start of my flash data leaving 512 K Bytes for program
  
   uint32_t myflash_erase_address;

   myflash_erase_address = FLASH_TARGET_OFFSET + (sector<<12);  // each sector is 4096 bytes

   ints = save_and_disable_interrupts();  // core 0 is using interrupts

   multicore_reset_core1();
   multicore_fifo_pop_blocking();  // returns a 0

   flash_range_erase(myflash_erase_address, FLASH_SECTOR_SIZE);

   restore_interrupts(ints);   // core 0 is using interrupts

   multicore_launch_core1(core1_entry);
   sleep_ms(10);   // Wait for it to start up    
}


For Reading, you Read a byte at a time. You do not need to disable core 1.

Reading is Referenced to the Flash Memory's address in the RP2040's address space and starts at 
XIP_BASE = 0x1000 0000 (4,294,967,296). That is where the program memory starts.

My logs start at 512 KBytes after that. Since I am using 1 MByte this gives me 512K Bytes left if I want to use it.

For reading my Flash data, in the RP2040's memory space
#define XIP_BASE 0x1000 0000     // already defined, for reading
#define FLASH_READ_ADDRESS (XIP_BASE + FLASH_TARGET_ADDRESS)

To access it by the start of the page (there are 4096 pages):

   uint32_t myflash_read_address;
   uint8_t *p;
   uint8_t data;

   myflash_read_address = XIP_BASE + (512 * 1024) + (page<<8);
   p = (uint8_t *)myflash_read_address;

   data = *p++;

Auto-incrementing the pointer makes it easy to grab as much datat as you want.

The SDK version I am using reports:
   Version: 1.95.1 (user setup)
   Commit: 65edc4939843c90c34d61f4ce11704f09d3e5cb6
   Date: 2024-10-31T05:14:54.222Z
   Electron: 32.2.1
   ElectronBuildId: 10427718
   Chromium: 128.0.6613.186
   Node.js: 20.18.0
   V8: 12.8.374.38-electron.0
   OS: Windows_NT x64 10.0.19045
*/
//***********************************************************************
